"""
Basic Node
"""
from __future__ import annotations
import uuid

from lattice.edge.edge import Edge


class Node:
    _successors: set
    _predecessors: set
    _label: str
    _fingerprint: uuid.UUID

    def __init__(self, pointer, label=None):
        """instantiate the node, if no label is given, then it is just the string representation of the data"""
        self.pointer = pointer
        self._predecessors = set()
        self._successors = set()
        self._label = label or str(pointer)
        self._fingerprint = uuid.uuid4()

    def copy(self, label=None):
        return Node(self.pointer, label)

    @property
    def data(self):
        return self.pointer

    @property
    def successors(self):
        return self._successors

    @property
    def predecessors(self):
        return self._predecessors

    def add_edge(self, edge: Edge, node:Node):
        self._add_successor(edge, node)
        node._add_predecessor(edge, self)

    def _add_successor(self, edge: Edge, node: Node):
        self._successors.add((edge, node))

    def _add_predecessor(self, edge: Edge, node: Node):
        self._predecessors.add((edge, node))

    def get_label(self):
        return self._label

    def __repr__(self):
        return str(self._fingerprint)



from lattice import Graph
from lattice import Node
from lattice import VisualEdge as Edge
import itertools

# base node setup
graph = Graph()
root_node = Node(None, label='0')
graph.add_nodes(**{'0': root_node})
vfront = [root_node]  # vertical frontier
hfront = [root_node]  # horizontal frontier

for i in range(1, 6):
    # refresh frontier
    new_vfront = []
    new_hfront = []

    for vnode in vfront:
        new_node = Node(None, label=str(i))
        vnode.add_edge(Edge(), new_node)
        new_vfront.append(new_node)

    for hnode in hfront:
        new_node = Node(None, label=str(i))
        hnode.add_edge(Edge(), new_node)
        new_hfront.append(new_node)

    corner_node = Node(None, label=str(i))
    new_vfront.append(corner_node)
    new_hfront.append(corner_node)

    for node1, node2 in itertools.pairwise(new_vfront):
        node1.add_edge(Edge(), node2)

    for node1, node2 in itertools.pairwise(new_hfront):
        node1.add_edge(Edge(), node2)

    vfront = new_vfront
    hfront = new_hfront

print(graph)



from .graph import Graph
from ..edge.edge import Edge


class Tree(Graph):
    """
    A tree is a type of graph that constrains relationship types
    Trees are formally constrained as graphs where all nodes except the root must have one predecessor
    """

    def __init__(self, root_node, copy=True):
        super().__init__()
        if copy:
            self.root_node = root_node.copy()
        else:
            self.root_node = root_node
        self._nodes.add(self.root_node)

    def add_nodes(self, parent_node, edge: Edge, **kwargs):
        """
        A node must be added by specifying an existing node as a predecessor
        """
        for key, value in kwargs.items():
            new_node = value.copy(key)
            setattr(self, key, new_node)

            self.add_edge(parent_node, new_node, edge)



"""
A balanced binary tree is a type of graph that constrains nodes to have at most 2 children
Both the left and right side of the tree must have the same depth
We can also restructure the tree to allow for O(log(n)) search
"""
from .tree import Tree
from ..edge.edge import Edge


class BBTree(Tree):
    """
    Edges are inconsequential
    TODO: Current implementation creates an unbalanced tree, will probably have to look into a more clever solution for this
    """
    _frontier = []  # frontier queue defines which leaves should be added to next

    def __init__(self, root_node):
        root_node = root_node.copy(label=root_node.get_label())
        super().__init__(root_node, copy=False)
        self._frontier.append(root_node)
        self._frontier.append(root_node)

    def add_nodes(self, *nodes):
        """
        Allows adding multiple nodes at a time
        """
        for given_node in nodes:
            self.add_node(given_node)

    def add_node(self, node, **kwargs):
        """
        Nodes can only be added to the leaves of the tree
        We define the leaves as the frontier that can be added to
        """
        new_node = node.copy(label=node.get_label())
        leaf_node = self._frontier.pop(0)
        self._nodes.add(new_node)

        # Add the node to the frontier twice as the node can have 2 children
        self._frontier.append(new_node)
        self._frontier.append(new_node)

        # Finally add an edge between the leaf node and the new node
        self.add_edge(leaf_node, new_node, Edge())



"""
The main focal data type of the project

NOTES:
    Edges have two options - Node.successor -> next node, edge
                             Node.successor -> edge -> next node
    We choose the first one as we can avoid the extra overhead and ensure
    That edges only exist within a graph

TODO:
- Find a more elegant way to reference nodes in a graph for operations such as finding edges
"""
import uuid
import os
from typing import List, Dict

import graphviz
from ..edge.visualedge import VisualEdge
from ..node.node import Node


class Graph:
    # A set of exposed nodes to access when they are called for (NOTE: may not include all nodes)
    _exposed_nodes: dict

    def __init__(self):
        self._nodes = {}  # initialise nodes as an empty set
        self.fingerprint = uuid.uuid4()

    @property
    def nodes(self) -> Dict[str, Node]:
        """Return the exposed nodes that belong to this graph """
        return self._nodes

    def add_nodes(self, **nodes: Node):
        """Exposed nodes are defined in key value pairs"""
        # add regular nodes
        for key, value in nodes.items():
            self._nodes[key] = value

    def add_node(self, node: Node, label):
        """Create a node within the graph"""
        self._nodes[label] = node

    def get_node(self, name: str):
        return self._nodes.get(name)

    def deepcopy(self):
        """Return a network of nodes encased by this graph, but copied"""
        new_graph = Graph()
        explored = {}
        for name, node in self.nodes.items():
            if node not in explored:
                new_node = node.copy(label=node.get_label())
                explored[node] = new_node
                new_graph.add_nodes(**{name: new_node})
                explored = self._recursive_copier(new_node, node, explored, new_graph)

        return new_graph

    def _recursive_copier(self, new_node, frontier_node, explored, graph):
        """Explore each new node and copy it into a given graph"""
        for edge, next_node in frontier_node.successors:
            if next_node not in explored:
                next_new_node = next_node.copy(label=next_node.get_label())
                explored[next_node] = next_new_node
                if next_node in self.nodes.values():
                    node_name = list(self._nodes.keys())[list(self._nodes.values()).index(next_node)]
                    graph.add_nodes(**{node_name: next_new_node})

                explored = self._recursive_copier(next_new_node, next_node, explored, graph)

            new_node.add_edge(edge, explored[next_node])
        return explored

    def visualise(self, fingerprint: str = None):
        """
        Display visible nodes in the graph
        Node names are set by the label of the node, rather than the key set by the graph
        """
        # Create a new fingerprint for each visualisation to generate a single file per print
        runtime_fingerprint = fingerprint or uuid.uuid4()
        dot = graphviz.Digraph(str(runtime_fingerprint))
        explored = []
        for name, node in self.nodes.items():
            if node not in explored:
                explored = self._recursive_visualiser(node, explored, dot)

        return dot

    def _recursive_visualiser(self, node, explored, dot):
        """Visualise every node by exploring node connections"""
        explored.append(node)
        shape = 'circle'
        if node in self._nodes.values():
            shape = 'doublecircle'
        dot.node(str(node), node.get_label(), shape=shape)
        for edge, next_node in node.successors:
            if isinstance(edge, VisualEdge):
                dot.edge(str(node), str(next_node), label=str(edge.get_label()))
                if next_node not in explored:
                    explored = self._recursive_visualiser(next_node, explored, dot)

        return explored

    def __repr__(self):
        graphoutput = self.visualise()
        dotoutput = graphoutput.pipe(format='dot', encoding='utf-8')
        os.system('echo \'' + dotoutput + '\' | graph-easy --from=dot --as_ascii')
        # graphoutput.render(view=True)
        return "​"



from lattice import Graph
from lattice import Node
from lattice import VisualEdge as Edge

grapha = Graph()
grapha.add_nodes(**{'0': Node(None, label='0')})
for i in range(1, 4):
    graphb = Graph()
    graphb.add_nodes(**{str(i): Node(None, label=str(i))})
    # create two copies of the original graph (they don't need to be added to the new graph)
    copya = Node(grapha.deepcopy())
    copyb = Node(grapha.deepcopy())

    # Add child edges
    graphb.get_node(str(i)).add_edge(Edge(), copya.data.get_node(str(i - 1)))
    graphb.get_node(str(i)).add_edge(Edge(), copyb.data.get_node(str(i - 1)))

    grapha = graphb

print(grapha)



"""
A type of node that requires enrichment of cost and label data
"""
from .edge import Edge


class VisualEdge(Edge):
    """
    This is a simple edge that can have a label attached
    It has no expansion on the regular edge as its only relevance is its existence
    To connect edges for a visual representation, this edge must be used
    """
    pass



"""
A type of node that requires enrichment of cost and label data
"""
from .edge import Edge


class TraverseEdge(Edge):
    """
    This edge requires that a label and a cost is assigned
    """

    def __init__(self, cost: int, label: str):
        super().__init__()
        self._data = cost
        self._label = label

    def get_cost(self) -> int:
        """Return the assigned cost of the edge"""
        return self.get_data()



"""
Edges are the connections between nodes
"""

from typing import Any


class Edge:
    """
    Default edge, requires a source and a destination
    We define a cost and a label, attempting to access them will return nothing
    """
    _data: Any
    _label: str

    def __init__(self, label=''):
        self._data = None
        self._label = label

    def get_data(self) -> Any:
        """Get the data we have stored in the edge"""
        return self._data

    def get_label(self) -> str:
        """Return the label assigned to this edge"""
        return self._label






"""
This file serves as an example of building graphs within the capabilities of the code we have
"""
from lattice import Node
from lattice import TraverseEdge
from lattice import Edge
from lattice import BBTree, Graph

# Balanced Binary Tree
def build_binary_tree(n):
    graph = BBTree(Node(None, label=str(0)))
    for i in range(1, n + 1):
        graph.add_node(Node(None, label=str(i)))

    return graph


graph2 = build_binary_tree(20)

if __name__ == "__main__":
    # We have created several graphs, uncomment lines here to visualise them
    # graph1.visualise()
    # graph2.visualise()
    # print(graph1)
    # print(graph2)
    pass



from lattice import Graph
from lattice import Node
from lattice import VisualEdge as Edge

graph = Graph()
graph.add_nodes(**{'1': Node(1), '2': Node(2), '3': Node(3)})
graph.get_node('1').add_edge(Edge(), graph.get_node('2'))
graph.get_node('2').add_edge(Edge(), graph.get_node('1'))
graph.get_node('2').add_edge(Edge('second'), graph.get_node('3'))


print(graph)
